/* eval3.c
	 5 Jan 2005
 */

#include "nut2.h"

int M[MEMEND];
int sp = STKBASE;	// M = heap : SS
int fp = STKBASE;

void push(exp a){
	sp++;
	if(sp >= MEMEND ) error("stack overflow");
	M[sp] = a;
}

exp head(exp e){
	if( e == NIL) return NIL;
	if( e < 0 || e >= MEMMAX )
		error("access memory out of bound");
	return M[e];
}

exp tail(exp e){
	if( e == NIL) return NIL;
	if( e < 0 || e >= MEMMAX )
		error("access memory out of bound");
	return M[e+1];
}

int isAtom(exp e){
	return (e & 0x80000000) != 0;
}

void decode(exp a, int *op, int *arg){
	*op = (a >> 24) & 0x07f;
	*arg = a & 0x0ffffff;
}

exp arg1(exp e){
	return head(e);
}

exp arg2(exp e){
	return head(tail(e));
}

exp arg3(exp e){
	return head(tail(tail(e)));
}

int cnt = 0;

exp eval(exp e){
	int op, arg, arity, k, v;
	int idx, fs, e1;

	if (cnt++ > 1000000)
		error("infinite loop\n");

	if( e == NIL) return NIL;
	if( ! isAtom(e) ) {
		e1 = tail(e);			// argument list (arg1 arg2 ...)
		e = head(e);			// operator
	}
	decode(e,&op,&arg);
	switch(op){
	case xIF:
		if( eval(arg1(e1)) != 0)
			return eval(arg2(e1));
		else
			return eval(arg3(e1));
	case xWHILE:
		while( eval(arg1(e1)) )
			v = eval(arg2(e1));
		return v;
	case xDO:
		while( e1 != NIL ){
			v = eval(head(e1));
			e1 = tail(e1);
		}
		return v;			// return the last eval
	case xNEW: return new(eval(e1));
	case xADD: return eval(arg1(e1)) + eval(arg2(e1));
	case xSUB: return eval(arg1(e1)) - eval(arg2(e1));
	case xEQ:  return eval(arg1(e1)) == eval(arg2(e1)) ? 1 : 0;
	case xLT: return eval(arg1(e1)) < eval(arg2(e1)) ? 1 : 0;
	case xGT: return eval(arg1(e1)) > eval(arg2(e1)) ? 1 : 0;
	case xCALL:
		while( e1 != NIL ){	// eval all arg
			v = eval(head(e1));
			push(v);
			e1 = tail(e1);
		}
		return eval(arg);	// eval body of fun
	case xLIT: return arg;
	case xSTR: return arg;
	case xGET: return M[fp-arg];
	case xPUT:
		v = eval(arg1(e1));
		M[fp-arg] = v;
		return v;
	case xLDX:
		idx = eval(arg1(e1));
		return M[ M[fp-arg] + idx];
	case xSTX:
		idx = eval(arg1(e1));
		v = eval(arg2(e1));
		M[ M[fp-arg] + idx] = v;
		return v;
	case xLDY:
		idx = eval(arg1(e1));
		return M[arg + idx];
	case xSTY:
		idx = eval(arg1(e1));
		v = eval(arg2(e1));
		M[arg + idx] = v;
		return v;
	case xLD: return M[arg]; break;
	case xST:
		v = eval(arg1(e1));
		M[arg] = v;
		return v;
	case xFUN:
		// fs, k should be compute by compiler
		// k = (arg & 0xff)-(arg >> 8)+1
	    // fs = (arg & 0xff)-1
//		arity = arg >> 8;
//		fs = arg & 0xff;
//		k = fs-arity+1;
		fs = arg & 0xff;
		k = arg >> 8;		// already recode
		M[sp+k] = fp;		// new frame
		fp = sp+k;
		sp = fp;
		v = eval(arg1(e1));
		sp = fp-fs;			// delete frame
		fp = M[fp];
		return v;
	case xSYS:
		v = eval(arg1(e1));
		switch(arg){
		case 1: printf("%d",v); break;
		case 2: printf("%c",v); break;
		}
		return NIL;
	default:
		printf("%d ",op);
		error("unknown op\n");
	}
	return NIL;
}

